from datasets import load_dataset
from torch.cuda import is_available

from ragas import evaluate
from ragas.metrics import answer_relevancy, context_relevancy, faithfulness

DEVICE = "cuda" if is_available() else "cpu"

# data
ds = load_dataset("explodinggradients/fiqa", "ragas_eval")["baseline"]

if __name__ == "__main__":
    result = evaluate(
        ds,
        metrics=[answer_relevancy, context_relevancy, faithfulness],
    )
    print(result)



from __future__ import annotations

import typing as t
from dataclasses import dataclass

import numpy as np
import torch
import transformers
from datasets import Dataset
from torch import Tensor
from torch.nn import CrossEntropyLoss
from torch.nn.functional import normalize
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import AutoConfig, AutoTokenizer
from transformers.models.auto.modeling_auto import MODEL_WITH_LM_HEAD_MAPPING_NAMES

from ragas.metrics.base import Metric

if t.TYPE_CHECKING:
    import numpy.typing as npt


class QGen:
    def __init__(self, model_name: str, device: str) -> None:
        config = AutoConfig.from_pretrained(model_name)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, model_max_length=512)
        if self.tokenizer.pad_token is None:
            self.tokenizer.pad_token = "[PAD]"
        architecture = np.intersect1d(
            list(MODEL_WITH_LM_HEAD_MAPPING_NAMES.values()), config.architectures
        )
        self.model = getattr(transformers, architecture[0]).from_pretrained(model_name)
        self.model.eval()

        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self._target_device = torch.device(device)

        self.model.to(device)
        self.max_len = 512
        self.question_template = "Generate a question for the given passage\n"
        self.loss_fun = CrossEntropyLoss(reduction="none")
        self.collate_fn = (
            self.encoder_decoder_collatefn
            if self.model.config.is_encoder_decoder
            else self.decoder_only_collatefn
        )

    def decoder_only_collatefn(self, batch):
        query, passage = zip(*batch)
        inputs = self.tokenizer(
            [self.question_template] * len(passage), passage
        ).input_ids
        labels = self.tokenizer(query).input_ids
        for i in range(len(batch)):
            if len(inputs[i] + labels[i]) > self.max_len:
                inputs[i] = inputs[i][: self.max_len]

        input_ids = [input + label for input, label in zip(inputs, labels)]
        batch_max_len = max(len(input) for input in input_ids)
        input_ids = self.tokenizer.pad(
            {"input_ids": input_ids}, max_length=batch_max_len, return_tensors="pt"
        )["input_ids"]
        mask_tensor = torch.zeros(*input_ids.shape)
        for i in range(input_ids.shape[0]):
            mask_tensor[i, : len(inputs[i])] = 1
            mask_tensor = torch.where(
                input_ids == self.tokenizer.pad_token_id, torch.tensor(1.0), mask_tensor
            )
        label_ids = torch.where(mask_tensor.bool(), torch.tensor(-100), input_ids)

        return {"input_ids": input_ids}, label_ids

    def encoder_decoder_collatefn(self, batch):
        query, passage = zip(*batch)
        inputs = self.tokenizer(
            [self.question_template] * len(passage),
            passage,
            max_length=self.max_len,
            padding=True,
            truncation=True,
            return_tensors="pt",
            add_special_tokens=True,
        )
        labels = self.tokenizer(
            query,
            max_length=self.max_len,
            truncation=True,
            padding=True,
            return_tensors="pt",
        ).input_ids
        labels = torch.where(
            labels == self.tokenizer.pad_token_id, torch.tensor(-100), labels
        )

        return {
            "input_ids": inputs.input_ids,
            "attention_mask": inputs.attention_mask,
            "labels": labels,
        }, labels

    def get_loss(self, logits: Tensor, labels: Tensor):
        if not self.model.config.is_encoder_decoder:
            logits = logits[:, :-1, :].contiguous()
            labels = labels[:, 1:].contiguous()
        losses = self.loss_fun(
            logits.view(-1, logits.shape[-1]), labels.view(-1)
        ).reshape(*labels.shape)
        losses = [
            1 - normalize(loss[loss != 0], dim=-1).mean(dim=-1).item()
            for loss in losses
        ]
        return np.round(losses, 3)

    def predict(
        self,
        sentences: list[list[str]],
        batch_size: int = 32,
        show_progress: bool = True,
    ) -> npt.NDArray[np.float64]:
        predictions = []
        dataloader = DataLoader(
            sentences, batch_size=batch_size, collate_fn=self.collate_fn  # type: ignore
        )

        if show_progress:
            dataloader = tqdm(dataloader)

        for _, data in enumerate(dataloader):
            inputs, labels = data
            inputs = {k: v.to(self._target_device) for k, v in inputs.items()}
            labels = labels.to(self._target_device)
            with torch.no_grad():
                logits = self.model(**inputs, output_hidden_states=False).logits
                loss = self.get_loss(logits, labels)
                predictions.append(loss)

        return np.hstack(predictions)


@dataclass
class AnswerRelevancy(Metric):
    batch_size: int = 32
    name: str = "answer_relevancy"
    model_name: str = "t5-base"

    def init_model(self: t.Self):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = QGen(self.model_name, self.device)

    @staticmethod
    def _make_question_answer_pairs(row: dict) -> dict:
        row["sentences"] = list(zip(row["question"], row["answer"]))
        return row

    def score(self: t.Self, dataset: Dataset) -> Dataset:
        """
        dataset: Dataset[question: list[str], answer: list[str]]
        """

        sentence_ds = dataset.map(
            self._make_question_answer_pairs, batched=True, batch_size=10
        )
        # we loose memory here because we have to make it py_list
        scores = self.model.predict(sentence_ds["sentences"])
        return Dataset.from_dict({f"{self.name}": scores})


answer_relevancy = AnswerRelevancy()



import pytest

from ragas.metrics.base import make_batches


@pytest.mark.parametrize(
    "batch_size, total_size, len_expected", [(5, 10, 2), (5, 11, 3), (5, 9, 2)]
)
def test_make_batches(batch_size, total_size, len_expected):
    batches = make_batches(total_size, batch_size)
    assert len(batches) == len_expected



from __future__ import annotations

import logging
import os
import typing as t
from functools import lru_cache
from warnings import warn

import torch
from torch import device as Device

DEVICES = ["cpu", "cuda"]
DEBUG_ENV_VAR = "RAGAS_DEBUG"


def device_check(device: t.Literal["cpu", "cuda"] | Device) -> torch.device:
    if isinstance(device, Device):
        return device
    if device not in DEVICES:
        raise ValueError(f"Invalid device {device}")
    if device == "cuda" and not torch.cuda.is_available():
        warn("cuda not available, using cpu")
        device = "cpu"

    return torch.device(device)


@lru_cache(maxsize=1)
def get_debug_mode() -> bool:
    if os.environ.get(DEBUG_ENV_VAR, str(False)).lower() == "true":
        logging.basicConfig(level=logging.DEBUG)
        return True
    else:
        return False



from __future__ import annotations

import typing as t
from dataclasses import dataclass
from itertools import combinations, product
from typing import List

import numpy as np
from datasets import Dataset
from langchain.chat_models.base import BaseChatModel
from langchain.llms.base import BaseLLM
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate
from sentence_transformers import CrossEncoder
from tqdm import tqdm

from ragas.metrics.base import Metric
from ragas.metrics.llms import generate

CONTEXT_RELEVANCE = HumanMessagePromptTemplate.from_template(
    """\
Task: Candidate sentence extraction.
Given the question and context, extract minimum number of sentences from context required to answer the question. If the context do not contain information required to answer the question return "No candidate sentences found".

question: Which equation is known as worlds most famous equation?
context:\nAlbert Einstein (14 March 1879 – 18 April 1955) was a German-born theoretical physicist,[5] widely ranked among the greatest and most influential scientists of all time. Best known for developing the theory of relativity, he also made important contributions to quantum mechanics, and was thus a central figure in the revolutionary reshaping of the scientific understanding of nature that modern physics accomplished in the first decades of the twentieth century.
His mass–energy equivalence formula E = mc2, which arises from relativity theory, has been called "the world's most famous equation".
sentences:His mass–energy equivalence formula E = mc2, which arises from relativity theory, has been called "the world's most famous equation".

question: Were Scott Derrickson and Ed Wood of the same nationality?
context :\nScott Derrickson (born July 16, 1966) is an American director, screenwriter and producer He lives in Los Angeles, California He is best known for directing horror films such as "Sinister", "The Exorcism of Emily Rose", and "Deliver Us From Evil", as well as the 2016 Marvel Cinematic Universe installment, "Doctor Strange"Tyler Bates is an American musician, music producer, and composer for films, television, and video games. Adam Collis is an American filmmaker and actor.Conrad Brooks is an American actor.Edward Davis Wood Jr. (October 10, 1924 – December 10, 1978) was an American filmmaker, actor, writer, producer, and director.
Now given a question and context, extract the minimum number of sentences from the given context required to answer the question completely. 
sentences:Scott Derrickson (born July 16, 1966) is an American director, screenwriter and producer. Edward Davis Wood Jr. (October 10, 1924 – December 10, 1978) was an American filmmaker, actor, writer, producer, and director.

question:{question}
context:\n{context}
sentences:"""  # noqa: E501
)


def sent_tokenize(sent: str) -> List[str]:
    return [s[:-1] if s.endswith(".") else s for s in sent.strip().split(". ")]


class SentenceAgreement:
    def __init__(
        self: t.Self,
        model_name: str = "cross-encoder/stsb-TinyBERT-L-4",
        metric: str = "bert_score",
    ):
        self.metric = metric
        self.cross_encoder = CrossEncoder(model_name)

    def bert_score(self, para1: str, para2: str) -> float:
        sentences1, sentences2 = sent_tokenize(para1), sent_tokenize(para2)
        scores = self.cross_encoder.predict(
            list(product(sentences1, sentences2)), convert_to_numpy=True  # type: ignore
        )
        assert isinstance(scores, np.ndarray), "Expects ndarray"
        scores = scores.reshape(len(sentences1), len(sentences2))
        return scores.max(axis=1).mean()

    @staticmethod
    def jaccard_score(para1: str, para2: str) -> float:
        sentences1, sentences2 = sent_tokenize(para1), sent_tokenize(para2)
        intersect = len(np.intersect1d(sentences1, sentences2))
        union = len(np.union1d(sentences1, sentences2))
        return intersect / union

    def evaluate(self, answers: List[str]) -> np.float_:
        """
        eval nC2 combinations
        """
        scores = []
        groups = combinations(answers, 2)
        for group in groups:
            if self.metric == "jaccard":
                score = self.jaccard_score(*group)  # type: ignore
            elif self.metric == "bert_score":
                score = self.bert_score(*group)  # type: ignore
            else:
                score = 0
                raise ValueError(f"Metric {self.metric} unavailable")
            scores.append(score)
        score = np.mean(scores)
        return score


@dataclass
class ContextRelevancy(Metric):
    """
    Extracts sentences from the context that are relevant to the question with
    self-consistancy checks. The number of relevant sentences and is used as the score.

    Attributes
    ----------
    name : str
    batch_size : int
        Batch size for openai completion.
    strictness : int
        Controls the number of times sentence extraction is performed to quantify
        uncertainty from the LLM. Defaults to 2.
    agreement_metric : str
        "bert_score" or "jaccard_score", used to measure agreement between multiple
        samples.
    model_name : str
        any encoder model. Used for calculating bert_score.
    """

    name: str = "context_relavency"
    batch_size: int = 15
    llm: t.Optional[BaseLLM | BaseChatModel] = None
    strictness: int = 2
    agreement_metric: str = "bert_score"
    model_name: str = "cross-encoder/stsb-TinyBERT-L-4"

    def __post_init__(self: t.Self):
        if self.agreement_metric == "bert_score" and self.model_name is None:
            raise ValueError(
                "model_name must be provided when agreement_metric is bert_score"
            )
        super().__post_init__()

    def init_model(self: t.Self):
        self.sent_agreement = SentenceAgreement(
            model_name=self.model_name, metric=self.agreement_metric
        )

    def score(self: t.Self, dataset: Dataset) -> Dataset:
        """
        Parameters
        ----------
        dataset: Dataset[question: list[str], contexts: list[list[str]]]

        Returns
        -------
        Dataset[question: list[str], contexts: list[list[str]], scores: list[float]]
            Dataset with the scores for each row.
        """
        if self.llm is None:
            raise ValueError("llm must not be None")
        prompts = []
        questions, contexts = dataset["question"], dataset["contexts"]
        for q, c in zip(questions, contexts):
            human_prompt = CONTEXT_RELEVANCE.format(question=q, context="\n".join(c))
            prompts.append(ChatPromptTemplate.from_messages([human_prompt]))

        responses: list[list[str]] = []
        for batch_idx in tqdm(range(0, len(prompts), 20)):
            results = generate(
                prompts[batch_idx : batch_idx + 20], self.llm, n=self.strictness
            )
            batch_responses = [[i.text for i in r] for r in results.generations]
            responses.extend(batch_responses)  # type: ignore

        scores = []
        for context, n_response in zip(contexts, responses):
            context = "\n".join(context)
            overlap_scores = []
            context_sents = sent_tokenize(context)
            for output in n_response:
                indices = [
                    context.find(sent)
                    for sent in sent_tokenize(output)
                    if context.find(sent) != -1
                ]
                overlap_scores.append(len(indices) / len(context_sents))
            if self.strictness > 1:
                agr_score = self.sent_agreement.evaluate(n_response)
            else:
                agr_score = 1
            scores.append(agr_score * np.mean(overlap_scores))

        return dataset.add_column(f"{self.name}", scores)  # type: ignore


context_relevancy = ContextRelevancy()



from __future__ import annotations

import logging
import os
import typing as t
from dataclasses import asdict, dataclass
from functools import lru_cache, wraps

import requests

from ragas.utils import get_debug_mode

if t.TYPE_CHECKING:
    P = t.ParamSpec("P")
    T = t.TypeVar("T")
    AsyncFunc = t.Callable[P, t.Coroutine[t.Any, t.Any, t.Any]]

logger = logging.getLogger(__name__)


USAGE_TRACKING_URL = "https://t.explodinggradients.com"
RAGAS_DO_NOT_TRACK = "RAGAS_DO_NOT_TRACK"
RAGAS_DEBUG_TRACKING = "__RAGAS_DEBUG_TRACKING"
USAGE_REQUESTS_TIMEOUT_SEC = 1


@lru_cache(maxsize=1)
def do_not_track() -> bool:  # pragma: no cover
    # Returns True if and only if the environment variable is defined and has value True
    # The function is cached for better performance.
    return os.environ.get(RAGAS_DO_NOT_TRACK, str(False)).lower() == "true"


@lru_cache(maxsize=1)
def _usage_event_debugging() -> bool:
    # For BentoML developers only - debug and print event payload if turned on
    return os.environ.get(RAGAS_DEBUG_TRACKING, str(False)).lower() == "true"


def silent(func: t.Callable[P, T]) -> t.Callable[P, T]:  # pragma: no cover
    # Silent errors when tracking
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> t.Any:
        try:
            return func(*args, **kwargs)
        except Exception as err:  # pylint: disable=broad-except
            if _usage_event_debugging():
                if get_debug_mode():
                    logger.error(
                        "Tracking Error: %s", err, stack_info=True, stacklevel=3
                    )
                else:
                    logger.info("Tracking Error: %s", err)
            else:
                logger.debug("Tracking Error: %s", err)

    return wrapper


@dataclass
class BaseEvent:
    event_type: str


@dataclass
class EvaluationEvent(BaseEvent):
    metrics: list[str]
    evaluation_mode: str
    num_rows: int


@silent
def track(event_properties: BaseEvent):
    if do_not_track():
        return

    payload = asdict(event_properties)

    if _usage_event_debugging():
        # For internal debugging purpose
        logger.info("Tracking Payload: %s", payload)
        return

    requests.post(USAGE_TRACKING_URL, json=payload, timeout=USAGE_REQUESTS_TIMEOUT_SEC)



from __future__ import annotations

import typing as t

from langchain.chat_models import ChatOpenAI
from langchain.chat_models.base import BaseChatModel
from langchain.llms import OpenAI
from langchain.llms.base import BaseLLM
from langchain.prompts import ChatPromptTemplate
from langchain.schema import LLMResult


def isOpenAI(llm: BaseLLM | BaseChatModel) -> bool:
    return isinstance(llm, OpenAI) or isinstance(llm, ChatOpenAI)


def generate(
    prompts: list[ChatPromptTemplate],
    llm: BaseLLM | BaseChatModel,
    n: t.Optional[int] = None,
) -> LLMResult:
    old_n = None
    if n is not None:
        if isinstance(llm, OpenAI) or isinstance(llm, ChatOpenAI):
            old_n = llm.n
            llm.n = n
        else:
            raise Exception(
                f"n={n} was passed to generate but the LLM {llm} does not support it."
                " Raise an issue if you want support for {llm}."
            )
    if isinstance(llm, BaseLLM):
        ps = [p.format() for p in prompts]
        result = llm.generate(ps)
    elif isinstance(llm, BaseChatModel):
        ps = [p.format_messages() for p in prompts]
        result = llm.generate(ps)

    if isinstance(llm, OpenAI) or isinstance(llm, ChatOpenAI):
        llm.n = old_n  # type: ignore
    return result



"""
Q - question
A - answer: generated_text from RAG pipeline
C - contexts: context used for generation
G - ground_truths: ground truth answer
"""
from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass
from math import floor

from datasets import Dataset
from langchain.chat_models.base import BaseChatModel
from langchain.llms.base import BaseLLM


def make_batches(total_size: int, batch_size: int) -> list[range]:
    """
    Take a total size and batch size and return a list of ranges for the batches
    """
    tail = total_size % batch_size
    num_batches = floor(total_size / batch_size)
    batches = [
        range(i, i + batch_size) for i in range(0, batch_size * num_batches, batch_size)
    ]
    if tail != 0:
        batches.append(range(batch_size * num_batches, batch_size * num_batches + tail))

    return batches


@dataclass
class Metric(ABC):
    batch_size: int
    llm: t.Optional[BaseLLM | BaseChatModel] = None

    def __post_init__(self: t.Self):
        if self.llm is None:
            from langchain.chat_models import ChatOpenAI

            self.llm = ChatOpenAI(model_name="gpt-3.5-turbo-16k")  # type: ignore

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def init_model():
        """
        This method will lazy initialize the model.
        """
        ...

    @abstractmethod
    def score(self: t.Self, dataset: Dataset) -> Dataset:
        ...

    def get_batches(self, dataset_size: int) -> list[range]:
        return make_batches(dataset_size, self.batch_size)



import concurrent.futures as f

from datasets import DatasetDict, load_dataset
from langchain.llms import OpenAI


def format_for_belar(row):
    row["context"] = row["selftext"]
    row["prompt"] = row["title"]
    row["ground_truth"] = row["answers"]["text"]
    return row


d = load_dataset("eli5")
assert isinstance(d, DatasetDict)
ds = d["test_eli5"].map(format_for_belar, batched=False)
ds = ds.select_columns(["context", "prompt", "ground_truth"])

ds = ds.shuffle(seed=42).select(range(500))
print(ds.shape, ds.column_names)


llm = OpenAI()  # type: ignore
prompt = """
{context}
with the above context explain like I'm five: {prompt}
"""


def get_answers(row):
    qs, cs = row["prompt"], row["context"]

    generated_answers = []
    with f.ThreadPoolExecutor(max_workers=10) as executor:
        results = executor.map(
            llm, [prompt.format(context=cs[i], prompt=qs[i]) for i in range(len(qs))]
        )
        for result in results:
            generated_answers.append(result)

    row["generated_answers"] = generated_answers
    return row


ds = ds.map(get_answers, batched=True, batch_size=10)



from __future__ import annotations

import time
import typing as t

import numpy as np
from rich.console import Console
from rich.table import Table

P = t.ParamSpec("P")
R = t.TypeVar("R")
OrigFunc = t.Callable[P, R]
DecoratedFunc = t.Callable[P, tuple[np.floating, np.floating]]


def timeit(func: OrigFunc, iteration: int = 3) -> DecoratedFunc:
    def function_timer(
        *args: P.args, **kwargs: P.kwargs
    ) -> tuple[np.floating, np.floating]:
        """
        Time the execution of a function and returns the time taken
        """
        # warmup
        func(*args, **kwargs)

        runtimes = []
        for _ in range(iteration):
            start = time.time()
            # we dont care about the return value
            func(*args, **kwargs)
            end = time.time()
            runtime = end - start
            runtimes.append(runtime)

        return np.mean(runtimes), np.var(runtimes)

    return function_timer


def print_table(result):
    table = Table("Batch Name", "(mean, var)", title="Benchmark Results")

    for batch_name, (mean, var) in result.items():
        table.add_row(batch_name, f"{mean:.4f}, {var:.4f}")

    console = Console()
    console.print(table)
